@@ -1,7 +1,7 @@
 /*
    Kickshaw - A Menu Editor for Openbox
 
-   Copyright (c) 2010–2022        Marcus Schätzle
+   Copyright (c) 2010–2024        Marcus Schätzle
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -19,6 +19,7 @@
 
 #include <gtk/gtk.h>
 
+#include <sys/stat.h>
 #include <stdlib.h> // Sometimes this might have to be included explicitly.
 #include <string.h> // Sometimes this might have to be included explicitly.
 
@@ -34,6 +35,30 @@
 */
 
 
+/* 
+
+    Version comparison function used in case that glibc is not installed.
+
+*/
+
+#if !__GLIBC__
+gint8 strverscmp_nonglibc (const gchar *version1, const gchar *version2) {
+    guint major1 = 0, minor1 = 0, bugfix1 = 0;
+    guint major2 = 0, minor2 = 0, bugfix2 = 0;
+
+    sscanf (version1, "%u.%u.%u", &major1, &minor1, &bugfix1);
+    sscanf (version2, "%u.%u.%u", &major2, &minor2, &bugfix2);
+
+    if (major1 < major2) return -1;
+    if (major1 > major2) return 1;
+    if (minor1 < minor2) return -1;
+    if (minor1 > minor2) return 1;
+    if (bugfix1 < bugfix2) return -1;
+    if (bugfix1 > bugfix2) return 1;
+    return 0;
+}
+#endif
+
 /* 
 
     Expands a row by retrieving the required path from a passed iterator.
@@ -47,30 +72,6 @@ void expand_row_from_iter (GtkTreeIter *local_iter)
     gtk_tree_view_expand_row (GTK_TREE_VIEW (ks.treeview), path, FALSE); // FALSE = just expand immediate children.
 }
 
-/*
-
-    Checks if a string contains a match defined by a regular expression.
-
-*/
-
-gboolean regex_match_simple (const gchar   *regex_str, 
-                             const gchar   *string, 
-                             const guint32  compile_options, 
-                             const guint32  match_options)
-{
-    gint error_number;
-    PCRE2_SIZE error_offset;
-
-    // error_number and error_offset are needed, even if they are not used.
-    g_autoptr(pcre2_code) compiled_pattern = pcre2_compile ((PCRE2_SPTR) regex_str, PCRE2_ZERO_TERMINATED, 
-                                                            compile_options, &error_number, &error_offset, NULL);
-    g_autoptr(pcre2_match_data) match_data = pcre2_match_data_create_from_pattern (compiled_pattern, NULL);
-
-    gint return_code = pcre2_match (compiled_pattern, (PCRE2_SPTR) string, PCRE2_ZERO_TERMINATED, 0, match_options, match_data, NULL);
-
-    return (return_code > 0);
-}
-
 /* 
 
    Extracts a substring from a string by using a regular expression.
@@ -79,23 +80,15 @@ gboolean regex_match_simple (const gchar   *regex_str,
 
 */
 
-gchar *extract_substring_via_regex (const gchar *string, 
-                                    const gchar *regex_str)
+gchar *extract_substring_via_regex (const gchar *string, const gchar *regex_str)
 {
+    // No compile and match options, no return location for a GError.
+    g_autoptr(GRegex) regex = g_regex_new (regex_str, 0, 0, NULL);
+    g_autoptr(GMatchInfo) match_info; // match_info is created even if g_regex_match returns FALSE.
     gchar *match;
-    gint error_number;
-    PCRE2_SIZE error_offset;
 
-    // error_number and error_offset are needed, even if they are not used.
-    g_autoptr(pcre2_code) compiled_pattern = pcre2_compile ((PCRE2_SPTR) regex_str, PCRE2_ZERO_TERMINATED, 
-                                                            PCRE2_UTF, &error_number, &error_offset, NULL);
-    g_autoptr(pcre2_match_data) match_data = pcre2_match_data_create_from_pattern (compiled_pattern, NULL);
-
-    pcre2_match (compiled_pattern, (PCRE2_SPTR) string, PCRE2_ZERO_TERMINATED, 0, 0, match_data, NULL);
-
-    PCRE2_SIZE *ovector = pcre2_get_ovector_pointer (match_data);    
-
-    match = g_strdup_printf ("%.*s", (gint) (ovector[1] - ovector[0]), (gchar *) string + ovector[0]);
+    g_regex_match (regex, string, 0, &match_info); // No match options.
+    match = g_match_info_fetch (match_info, 0); // 0 = full text of the match.
 
     return match;
 }
@@ -173,8 +166,9 @@ gchar *get_modification_time_for_icon (const gchar *icon_path)
     g_autoptr(GFileInfo) file_info = g_file_query_info (file, G_FILE_ATTRIBUTE_TIME_MODIFIED, G_FILE_QUERY_INFO_NONE, NULL, &error);
 
     if (G_UNLIKELY (error)) {
-        g_autofree gchar *err_msg = g_strdup_printf ("<b>Could not retrieve the modification time for the icon</b> <tt>%s</tt> <b>!\n"
-                                                     "Error:</b> %s", icon_path, error->message);
+        g_autofree gchar *err_msg = g_strdup_printf (_("<b>Could not retrieve the modification time for the icon</b> <tt>%s</tt> <b>!\n"
+                                                       "Error:</b> %s"), 
+                                                     icon_path, error->message);
 
         show_errmsg (err_msg);
     }
@@ -317,42 +311,6 @@ gchar *get_txt_from_resources (      GResource *resources,
 
 */
 
-/*
-
-	Determines which compile options to use according to the options set by the user.
-
-*/
-
-guint32 get_compile_options (const gboolean find_and_replace)
-{
-	gboolean regular_expression, match_case;
-	guint32 compile_options;
-
-	if (!find_and_replace) {
-		regular_expression = (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (ks.special_options[REGULAR_EXPRESSION])));
-		match_case = (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (ks.special_options[MATCH_CASE])));
-	}
-	else {
-		regular_expression = (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (ks.fr_special_options[REGULAR_EXPRESSION])));
-		match_case = (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (ks.fr_special_options[MATCH_CASE])));
-	}
-
-	if (regular_expression && !match_case) {
-		compile_options = PCRE2_UTF | PCRE2_CASELESS;
-    }
-	else if (regular_expression && match_case) {
-		compile_options = PCRE2_UTF;
-	}
-    else if (!regular_expression && !match_case) {
-		compile_options = PCRE2_UTF | PCRE2_LITERAL | PCRE2_CASELESS;
-	}
-	else { // !regular_expression && match_case
-		compile_options = PCRE2_UTF | PCRE2_LITERAL;
-	}
-
-	return compile_options;
-}
-
 /*
 
     A save way to set the values of several bits of the settings bit field
@@ -413,9 +371,10 @@ void delete_undo_stack_items (void)
         g_autoptr(GFile) undo_stack_item = g_file_new_for_path (undo_stack_item_str);
 
         if (G_UNLIKELY (!(g_file_delete (undo_stack_item, NULL, &error)))) {
-            g_autofree gchar *err_msg = g_strdup_printf ("<b>Could not remove undo stack item</b> <tt>%i</tt> "
-                                                         "<b>inside tmp folder</b> <tt>%s</tt> <b>!\n"
-                                                         "Error:</b> %s", stack_file_cnt, ks.tmp_path, error->message);
+            g_autofree gchar *err_msg = g_strdup_printf (_("<b>Could not remove undo stack item</b> <tt>%1$i</tt> "
+                                                           "<b>inside tmp folder</b> <tt>%2$s</tt> <b>!\n"
+                                                           "Error:</b> %3$s"), 
+                                                         stack_file_cnt, ks.tmp_path, error->message);
 
             show_errmsg (err_msg);
 
@@ -432,7 +391,7 @@ void delete_undo_stack_items (void)
 
 void delete_autosave (void)
 {
-    g_autofree gchar *autosave_file_path = g_build_filename (g_get_home_dir (), ".kickshaw_autosave", NULL);
+    g_autofree gchar *autosave_file_path = g_build_filename (ks.home_dir, ".kickshaw_autosave", NULL);
     g_autoptr(GFile) autosave_file = g_file_new_for_path (autosave_file_path);
 
     /* The autosave file does not exist after ...
@@ -444,7 +403,7 @@ void delete_autosave (void)
         g_file_delete (autosave_file, NULL, &error);
 
         if (G_UNLIKELY (error)) {
-            g_autofree gchar *err_msg = g_strdup_printf ("<b>Could not delete the autosave file</b> <tt>%s</tt> <b>!\nError:</b> %s", 
+            g_autofree gchar *err_msg = g_strdup_printf (_("<b>Could not delete the autosave file</b> <tt>%s</tt> <b>!\nError:</b> %s"), 
                                                          autosave_file_path, error->message);
 
             show_errmsg (err_msg);
@@ -493,7 +452,7 @@ GtkWidget *new_label_with_formattings (const gchar    *label_txt,
     The dialog can have up to three buttons. If less than three are needed, 
     the "button_txt_2" and/or "button_txt_3" argument should be set to NULL.
     If something else than a label text shall be added to the content area, 
-    the "show_immediately" argument should be set to FALSE. 
+    the "show_immediately" argument should be set to DONT_SHOW_IMMEDIATELY (=FALSE). 
     gtk_widget_show_all () should be called then after all additional widgets have been added to the content area.
 
 */
@@ -505,14 +464,12 @@ GtkWidget *create_dialog (      GtkWidget **dialog,
                           const gchar      *button_txt_1, 
                           const gchar      *button_txt_2, 
                           const gchar      *button_txt_3, 
-                          const gboolean    use_header_bar, 
                           const gboolean    show_immediately)
 {
     GtkWidget *content_area, *label;
     g_autofree gchar *label_txt_with_addl_border = g_strdup_printf ("\n%s\n", label_txt);
 
-    *dialog = gtk_dialog_new_with_buttons (NULL, GTK_WINDOW (ks.window), 
-                                           (use_header_bar) ? GTK_DIALOG_MODAL | GTK_DIALOG_USE_HEADER_BAR : GTK_DIALOG_MODAL, 
+    *dialog = gtk_dialog_new_with_buttons (NULL, GTK_WINDOW (ks.window), GTK_DIALOG_MODAL, 
                                            button_txt_1, 1, button_txt_2, 2, button_txt_3, 3,
                                            NULL);
 
@@ -570,8 +527,7 @@ void show_message_dialog (const gchar *dialog_title,
 {
     GtkWidget *dialog;
 
-	// FALSE = show header bar, TRUE = dialog is shown immediately
-    create_dialog (&dialog, dialog_title, icon_name, label_txt, button_txt, NULL, NULL, FALSE, TRUE);
+    create_dialog (&dialog, dialog_title, icon_name, label_txt, button_txt, NULL, NULL, SHOW_IMMEDIATELY);
 
     gtk_dialog_run (GTK_DIALOG (dialog));
     gtk_widget_destroy (dialog);
@@ -590,11 +546,11 @@ void create_file_dialog (      GtkWidget **dialog,
     g_autoptr(GFile) menu_folder = g_file_new_for_path (menu_folder_str);
     GtkFileFilter *file_filter;
 
-    *dialog = gtk_file_chooser_dialog_new ((open) ? "Open Openbox Menu File" : "Save Openbox Menu File As…", 
+    *dialog = gtk_file_chooser_dialog_new ((open) ? _("Open Openbox Menu File") : _("Save Openbox Menu File As…"), 
                                            GTK_WINDOW (ks.window),
                                            (open) ? GTK_FILE_CHOOSER_ACTION_OPEN : GTK_FILE_CHOOSER_ACTION_SAVE, 
-                                           "_Cancel", GTK_RESPONSE_CANCEL,
-                                           (open) ? "_Open" : "_Save", GTK_RESPONSE_ACCEPT,
+                                           C_("Cancel|File Dialogue", "_Cancel"), GTK_RESPONSE_CANCEL,
+                                           (open) ? _("_Open") : _("_Save"), GTK_RESPONSE_ACCEPT,
                                            NULL);
 
     if (g_file_query_exists (menu_folder, NULL)) {
@@ -602,7 +558,7 @@ void create_file_dialog (      GtkWidget **dialog,
     }
 
     file_filter = gtk_file_filter_new ();
-    gtk_file_filter_set_name (file_filter, "Openbox Menu Files");
+    gtk_file_filter_set_name (file_filter, _("Openbox Menu Files"));
     gtk_file_filter_add_pattern (file_filter, "*.xml");
     gtk_file_chooser_add_filter (GTK_FILE_CHOOSER (*dialog), file_filter);
 }
@@ -650,25 +606,22 @@ void determine_theme_type (void)
     if (at_startup || !STREQ (ks.red_hue, old_color)) {
         ks.blue_hue = (ks.bg_image_or_gradient) ? "blue" : ((!ks.dark_theme) ? "darkblue" : "lightblue");
 
-        g_string_printf (ks.label_field_txt, " Label (<span foreground='%s'>*</span>): ", ks.red_hue);
+        g_string_printf (ks.label_field_txt, _(" Label (<span foreground='%s'>*</span>): "), ks.red_hue);
         if (G_UNLIKELY (g_slist_find_custom (ks.menu_ids, "", (GCompareFunc) strcmp))) {
-            g_string_printf (ks.menu_ID_field_txt, " Menu ID (<span foreground='%s'>*</span>): ", ks.blue_hue);
+            g_string_printf (ks.menu_ID_field_txt, _(" Menu ID (<span foreground='%s'>*</span>): "), ks.blue_hue);
         }
         else {
-            g_string_assign (ks.menu_ID_field_txt, " Menu ID: ");
+            g_string_assign (ks.menu_ID_field_txt, _(" Menu ID: "));
         }
         g_string_printf (ks.execute_field_txt, " Execute (<span foreground='%s'>*</span>): ", ks.red_hue);
         g_string_printf (ks.prompt_field_txt, " Prompt (<span foreground='%s'>*</span>): ", ks.red_hue);
         g_string_printf (ks.command_field_txt, " Command (<span foreground='%s'>*</span>): ", ks.red_hue);
-        g_string_printf (ks.mandatory_label_txt, "(<span foreground='%s'>*</span>) = mandatory", ks.red_hue);
-        g_string_printf (ks.mandatory_empty_string_txt, "(<span foreground='%s'>*</span>) = mandatory; "
-                                                        "(<span foreground='%s'>*</span>) = "
-                                                        "mand. as empty string menu ID exists",
-                                                        ks.red_hue, ks.blue_hue);
-        g_string_printf (ks.double_menu_id_label_txt, " <span foreground='%s'>!!</span> "
-                                                      "Menu ID already exists "
-                                                      "<span foreground='%s'>!!</span> ", 
-                                                      ks.red_hue, ks.red_hue);
+        g_string_printf (ks.mandatory_label_txt, "(<span foreground='%s'>*</span>) = %s", ks.red_hue, _("mandatory"));
+        g_string_printf (ks.mandatory_empty_string_txt, "(<span foreground='%s'>*</span>) = %s; "
+                                                        "(<span foreground='%s'>*</span>) = %s",
+                                                        ks.red_hue, _("mandatory"), ks.blue_hue, _("mand. as empty-string menu ID exists"));
+        g_string_printf (ks.double_menu_id_label_txt, " <span foreground='%s'>!!</span> %s <span foreground='%s'>!!</span> ", 
+                                                      ks.red_hue,  _("Menu ID already exists "), ks.red_hue);
     }
 
     if (!at_startup && !STREQ (ks.red_hue, old_color)) {
@@ -864,7 +817,7 @@ guint8 check_if_invisible_ancestor_exists (GtkTreeModel *local_model,
 
         if (G_UNLIKELY (element_ancestor_visibility_txt_loop && 
                         g_str_has_prefix (element_ancestor_visibility_txt_loop, "invisible"))) {
-            visibility_of_ancestor = (g_str_has_suffix (element_ancestor_visibility_txt_loop, "orphaned menu")) ? 
+            visibility_of_ancestor = (g_str_has_suffix  (element_ancestor_visibility_txt_loop, "orphaned menu")) ? 
                                       INVISIBLE_ORPHANED_ANCESTOR : INVISIBLE_ANCESTOR;
 
             return visibility_of_ancestor;
@@ -892,7 +845,7 @@ gboolean add_icon_occurrence_to_list (G_GNUC_UNUSED GtkTreeModel *foreach_model,
         ks.rows_with_icons = g_slist_prepend (ks.rows_with_icons, gtk_tree_path_copy (foreach_path));
     }
 
-    return FALSE; // Continue iterating
+    return CONTINUE_ITERATING;
 }
 
 /* 
@@ -941,30 +894,23 @@ void find_in_columns_management (const gboolean    origin_is_fr,
 
 */
 
-gboolean check_if_regex_is_valid (GtkWidget *regex_checkbox, 
-                                  GtkWidget *entry_field)
+gboolean check_if_regex_is_valid (GtkWidget *regex_checkbox, GtkWidget *entry_field)
 {
     if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (regex_checkbox))) {
-        gint error_number;
-        PCRE2_SIZE error_offset;
+        g_autoptr(GError) error = NULL;
+        g_autoptr(GRegex) regex = g_regex_new (gtk_entry_get_text (GTK_ENTRY (entry_field)), 0, 0, &error);
 
-        // error_number and error_offset are needed, even if they are not used.
-        g_autoptr(pcre2_code) compiled_pattern = pcre2_compile ((PCRE2_SPTR) gtk_entry_get_text (GTK_ENTRY (entry_field)), 
-                                                                PCRE2_ZERO_TERMINATED, PCRE2_UTF, &error_number, &error_offset, NULL);
+        if (G_UNLIKELY (error)) {
+            g_autofree gchar *err_msg = g_strdup_printf (_("<b>The regular expression that has been entered is invalid.\n"
+                                                           "Error:</b> %s"), error->message);
 
-        if (!compiled_pattern) {
-            PCRE2_UCHAR buffer[256];
-            pcre2_get_error_message (error_number, buffer, sizeof (buffer));
-
-            g_autofree gchar *err_msg = g_strdup_printf ("PCRE2 compilation failed at offset %d: %s\n", (gint) error_offset, buffer);
-
-            show_errmsg (err_msg);   
+            show_errmsg (err_msg);           
 
             return FALSE;
         }
     }
 
-    return TRUE;        
+    return TRUE;
 }
 
 /* 
@@ -987,3 +933,31 @@ void visually_indicate_request_for_user_intervention (GtkWidget      *widget,
         -1, NULL);
     }
 }
+
+/*
+
+    Checks if Kickshaw is installed somewhere and returns the command suiting for a restart.
+
+*/
+
+const gchar *installed_or_local (void)
+{
+    g_auto(GStrv) paths = g_strsplit (g_getenv ("PATH"), ":", -1);
+    guint paths_idx = 0;
+    struct stat sb;
+    const gchar *command = "./kickshaw"; // Default
+
+    while (paths[paths_idx]) {
+        g_autofree gchar *path_with_iconv = g_strconcat (paths[paths_idx], "/kickshaw", NULL);
+
+        if (stat (path_with_iconv, &sb) == 0) {
+            command = "kickshaw";
+
+            break;
+        }
+
+        ++paths_idx;
+    }
+
+    return command;
+}
